/* Copyright (C) 2009-2021 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Valentina Akishina, Ivan Kisel, Sergey Gorbunov [committer], Igor Kulakov, Maksym Zyzak */

/*
 *=====================================================
 *
 *  CBM Level 1 4D Reconstruction
 *
 *  Authors: V.Akishina, I.Kisel,  S.Gorbunov, I.Kulakov, M.Zyzak
 *  Documentation: V.Akishina
 *
 *  e-mail : v.akishina@gsi.de
 *
 *=====================================================
 *
 *  Finds tracks using the Cellular Automaton algorithm
 *
 */


#include "CaTrackFinderWindow.h"

#include <algorithm>
#include <fstream>
#include <iostream>
#include <list>
#include <map>

#include <stdio.h>

#include "CaBranch.h"
#include "CaFramework.h"
#include "CaGrid.h"
#include "CaGridEntry.h"
#include "CaTrack.h"
#include "CaTrackParam.h"
#include "CaTripletConstructor.h"

// #include "CaToolsDebugger.h"


// using namespace std;
using std::cout;
using std::endl;
using Track = cbm::algo::ca::Track;

using namespace cbm::algo::ca;

// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinderWindow::TrackFinderWindow(ca::Framework& algo) : frAlgo(algo) {}

// ---------------------------------------------------------------------------------------------------------------------
//
TrackFinderWindow::~TrackFinderWindow() {}

// ---------------------------------------------------------------------------------------------------------------------
//


bool TrackFinderWindow::checkTripletMatch(const ca::Triplet& l, const ca::Triplet& r, fscal& dchi2) const
{
  dchi2 = 1.e20;

  if (r.GetMHit() != l.GetRHit()) return false;
  if (r.GetLHit() != l.GetMHit()) return false;

  if (r.GetMSta() != l.GetRSta()) return false;
  if (r.GetLSta() != l.GetMSta()) return false;


  if (r.IsMomentumFitted()) {
    assert(l.IsMomentumFitted());

    fscal dqp = l.GetQp() - r.GetQp();
    fscal Cqp = l.GetCqp() + r.GetCqp();

    if (!std::isfinite(dqp)) return false;
    if (!std::isfinite(Cqp)) return false;

    if (dqp * dqp > frAlgo.fTripletLinkChi2 * Cqp) {
      return false;  // bad neighbour // CHECKME why do we need recheck it?? (it really change result)
    }
    dchi2 = dqp * dqp / Cqp;
  }
  else {
    fscal dtx = l.GetTx() - r.GetTx();
    fscal Ctx = l.GetCtx() + r.GetCtx();

    fscal dty = l.GetTy() - r.GetTy();
    fscal Cty = l.GetCty() + r.GetCty();

    // it shouldn't happen, but happens sometimes

    if (!std::isfinite(dtx)) return false;
    if (!std::isfinite(dty)) return false;
    if (!std::isfinite(Ctx)) return false;
    if (!std::isfinite(Cty)) return false;

    if (dty * dty > frAlgo.fTripletLinkChi2 * Cty) return false;
    if (dtx * dtx > frAlgo.fTripletLinkChi2 * Ctx) return false;

    //dchi2 = 0.5f * (dtx * dtx / Ctx + dty * dty / Cty);
    dchi2 = 0.;
  }

  if (!std::isfinite(dchi2)) return false;

  return true;
}


// **************************************************************************************************
// *                                                                                                *
// *                            ------ CATrackFinder procedure ------                               *
// *                                                                                                *
// **************************************************************************************************

// ---------------------------------------------------------------------------------------------------------------------
//
void TrackFinderWindow::ReadWindowData()
{
  int nHits = 0;
  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); iS++) {
    frAlgo.fStationHitsStartIndex[iS] = nHits;
    frAlgo.fStationNhits[iS]          = frAlgo.fSliceHitIds[iS].size();
    nHits += frAlgo.fStationNhits[iS];
  }
  frAlgo.fStationHitsStartIndex[frAlgo.GetParameters().GetNstationsActive()] = nHits;

  frAlgo.fWindowHits.reset(nHits);
  frAlgo.fIsWindowHitSuppressed.reset(nHits, 0);

  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); iS++) {
    for (ca::HitIndex_t ih = 0; ih < frAlgo.fSliceHitIds[iS].size(); ++ih) {
      ca::Hit h = frAlgo.fInputData.GetHit(frAlgo.fSliceHitIds[iS][ih]);
      h.SetId(frAlgo.fSliceHitIds[iS][ih]);

      frAlgo.fWindowHits[frAlgo.fStationHitsStartIndex[iS] + ih] = h;
    }
  }

  frAlgo.fHitFirstTriplet.reset(nHits, 0);
  frAlgo.fHitNtriplets.reset(nHits, 0);


  frAlgo.fSliceRecoTracks.clear();
  frAlgo.fSliceRecoTracks.reserve(2 * nHits / frAlgo.GetParameters().GetNstationsActive());

  frAlgo.fSliceRecoHits.clear();
  frAlgo.fSliceRecoHits.reserve(2 * nHits);

  frAlgo.fTrackCandidates.clear();
  frAlgo.fTrackCandidates.reserve(nHits / 10);
  for (unsigned int iS = 0; iS < constants::size::MaxNstations; iS++) {
    int nHitsStation = frAlgo.fSliceHitIds[iS].size();
    frAlgo.fTriplets[iS].clear();
    frAlgo.fTriplets[iS].reserve(2 * nHitsStation);
  }
}

void TrackFinderWindow::CaTrackFinderSlice()
{
  frAlgo.fNFindIterations = frAlgo.GetParameters().GetNcaIterations();


  /********************************/ /**
   * CATrackFinder routine setting
   ***********************************/

  ReadWindowData();

  for (int iS = 0; iS < frAlgo.GetParameters().GetNstationsActive(); ++iS) {

    bool timeUninitialised = 1;
    fscal lasttime         = 0;
    fscal starttime        = 0;


    fscal gridMinX = -0.1;
    fscal gridMaxX = 0.1;
    fscal gridMinY = -0.1;
    fscal gridMaxY = 0.1;

    for (ca::HitIndex_t ih = 0; ih < frAlgo.fSliceHitIds[iS].size(); ++ih) {
      const ca::Hit& h = frAlgo.fInputData.GetHit(frAlgo.fSliceHitIds[iS][ih]);

      if (h.X() < gridMinX) { gridMinX = h.X(); }
      if (h.X() > gridMaxX) { gridMaxX = h.X(); }
      if (h.Y() < gridMinY) { gridMinY = h.Y(); }
      if (h.Y() > gridMaxY) { gridMaxY = h.Y(); }

      const fscal time = h.T();
      assert(std::isfinite(time));
      if (timeUninitialised || lasttime < time) { lasttime = time; }
      if (timeUninitialised || starttime > time) { starttime = time; }
      timeUninitialised = false;
    }

    // TODO: changing the grid also changes the result. Investigate why it happens.
    // TODO: find the optimal grid size

    int nSliceHits = frAlgo.fSliceHitIds[iS].size();
    fscal sizeY    = gridMaxY - gridMinY;
    fscal sizeX    = gridMaxX - gridMinX;
    int nBins2D    = 1 + nSliceHits;
    fscal yStep    = sizeY / sqrt(nBins2D);
    fscal xStep    = sizeX / sqrt(nBins2D);

    fscal scale = frAlgo.GetParameters().GetStation(iS).GetZScal() - frAlgo.GetParameters().GetTargetPositionZ()[0];

    if (yStep > 0.3 * scale) yStep = 0.3 * scale;
    if (yStep < 0.01 * scale) yStep = 0.01 * scale;
    if (xStep > 0.3 * scale) xStep = 0.3 * scale;
    if (xStep < 0.01 * scale) xStep = 0.01 * scale;

    frAlgo.vGrid[iS].BuildBins(gridMinX, gridMaxX, gridMinY, gridMaxY, xStep, yStep);
    frAlgo.vGrid[iS].StoreHits(frAlgo.fWindowHits, frAlgo.fStationHitsStartIndex[iS], frAlgo.fStationNhits[iS],
                               frAlgo.fvHitKeyFlags);
  }


  // ---- Loop over Track Finder iterations ----------------------------------------------------------------//


  assert(frAlgo.fNFindIterations == (int) frAlgo.GetParameters().GetCAIterations().size());

  for (frAlgo.fCurrentIterationIndex = 0;
       frAlgo.fCurrentIterationIndex < frAlgo.GetParameters().GetCAIterations().size();
       ++frAlgo.fCurrentIterationIndex) {

    // current iteration of the CA tracker
    const auto& caIteration   = frAlgo.GetParameters().GetCAIterations()[frAlgo.fCurrentIterationIndex];
    frAlgo.fpCurrentIteration = &caIteration;

    if (frAlgo.fCurrentIterationIndex > 0) {
      for (int ista = 0; ista < frAlgo.GetParameters().GetNstationsActive(); ++ista) {
        frAlgo.vGrid[ista].RemoveUsedHits(frAlgo.fWindowHits, frAlgo.fvHitKeyFlags);
      }
    }

    frAlgo.fIsWindowHitSuppressed.reset(frAlgo.fWindowHits.size(), 0);

    for (int j = 0; j < frAlgo.GetParameters().GetNstationsActive(); j++) {
      frAlgo.fTriplets[j].clear();
    }

    frAlgo.fHitFirstTriplet.reset(frAlgo.fWindowHits.size(), 0);
    frAlgo.fHitNtriplets.reset(frAlgo.fWindowHits.size(), 0);

    {
      // #pragma omp  task
      {
        // --- SET PARAMETERS FOR THE ITERATION ---

        frAlgo.fFirstCAstation      = caIteration.GetFirstStationIndex();
        frAlgo.fTrackChi2Cut        = caIteration.GetTrackChi2Cut();
        frAlgo.fDoubletChi2Cut      = caIteration.GetDoubletChi2Cut();  //11.3449 * 2.f / 3.f;  // prob = 0.1
        frAlgo.fTripletChi2Cut      = caIteration.GetTripletChi2Cut();  //21.1075;  // prob = 0.01%
        frAlgo.fTripletFinalChi2Cut = caIteration.GetTripletFinalChi2Cut();
        frAlgo.fPickGather          = caIteration.GetPickGather();       //3.0;
        frAlgo.fTripletLinkChi2     = caIteration.GetTripletLinkChi2();  //5.0;
        frAlgo.fMaxInvMom           = caIteration.GetMaxQp();            //1.0 / 0.5;  // max considered q/p
        frAlgo.fMaxSlopePV          = caIteration.GetMaxSlopePV();       //1.1;
        frAlgo.fMaxSlope            = caIteration.GetMaxSlope();         //2.748;  // corresponds to 70 grad

        // define the target
        frAlgo.fTargX = frAlgo.GetParameters().GetTargetPositionX();
        frAlgo.fTargY = frAlgo.GetParameters().GetTargetPositionY();
        frAlgo.fTargZ = frAlgo.GetParameters().GetTargetPositionZ();

        fscal SigmaTargetX = caIteration.GetTargetPosSigmaX();
        fscal SigmaTargetY = caIteration.GetTargetPosSigmaY();  // target constraint [cm]

        // Select magnetic field. For primary tracks - fVtxFieldValue, for secondary tracks - st.fieldSlice
        if (caIteration.GetPrimaryFlag()) { frAlgo.fTargB = frAlgo.GetParameters().GetVertexFieldValue(); }
        else {
          frAlgo.fTargB = frAlgo.GetParameters().GetStation(0).fieldSlice.GetFieldValue(0, 0);
        }  // NOTE: calculates field frAlgo.fTargB in the center of 0th station

        frAlgo.fIsTargetField = (fabs(frAlgo.fTargB.y[0]) > 0.001);

        frAlgo.TargetMeasurement.SetX(frAlgo.fTargX);
        frAlgo.TargetMeasurement.SetY(frAlgo.fTargY);
        frAlgo.TargetMeasurement.SetDx2(SigmaTargetX * SigmaTargetX);
        frAlgo.TargetMeasurement.SetDxy(0);
        frAlgo.TargetMeasurement.SetDy2(SigmaTargetY * SigmaTargetY);
        frAlgo.TargetMeasurement.SetNdfX(1);
        frAlgo.TargetMeasurement.SetNdfY(1);

        /// Set correction in order to take into account overlaping and iff z.
        /// The reason is that low momentum tracks are too curved and goes not from target direction. That's why sort by hit_y/hit_z is not work idealy
        /// If sort by y then it is max diff between same station's modules (~0.4cm)
        frAlgo.fMaxDZ = caIteration.GetMaxDZ();  //0;
      }
    }


    ///   stage for triplets creation

    frAlgo.fMonitorData.StartTimer(ETimer::TripletConstruction);

    ca::TripletConstructor constructor1(&frAlgo);
    ca::TripletConstructor constructor2(&frAlgo);
    ca::TripletConstructor constructor3(&frAlgo);

    for (int istal = frAlgo.GetParameters().GetNstationsActive() - 2; istal >= frAlgo.fFirstCAstation;
         istal--) {  // start with downstream chambers
      Tindex nGridEntriesL = frAlgo.vGrid[istal].GetEntries().size();

      for (Tindex iel = 0; iel < nGridEntriesL; ++iel) {
        ca::HitIndex_t ihitl = frAlgo.vGrid[istal].GetEntries()[iel].GetObjectId();
        constructor1.CreateTripletsForHit(istal, istal + 1, istal + 2, ihitl);

        if (frAlgo.fpCurrentIteration->GetJumpedFlag()) {
          constructor2.CreateTripletsForHit(istal, istal + 1, istal + 3, ihitl);
          constructor3.CreateTripletsForHit(istal, istal + 2, istal + 3, ihitl);
        }

        auto oldSize = frAlgo.fTriplets[istal].size();

        frAlgo.fTriplets[istal].insert(frAlgo.fTriplets[istal].end(), constructor1.GetTriplets().begin(),
                                       constructor1.GetTriplets().end());
        frAlgo.fTriplets[istal].insert(frAlgo.fTriplets[istal].end(), constructor2.GetTriplets().begin(),
                                       constructor2.GetTriplets().end());
        frAlgo.fTriplets[istal].insert(frAlgo.fTriplets[istal].end(), constructor3.GetTriplets().begin(),
                                       constructor3.GetTriplets().end());

        frAlgo.fHitFirstTriplet[ihitl] = frAlgo.PackTripletId(istal, oldSize);
        frAlgo.fHitNtriplets[ihitl]    = frAlgo.fTriplets[istal].size() - oldSize;
      }
    }  // istal

    frAlgo.fMonitorData.StopTimer(ETimer::TripletConstruction);

    // search for neighbouring triplets
    frAlgo.fMonitorData.StartTimer(ETimer::NeighboringTripletSearch);
    for (int istal = frAlgo.GetParameters().GetNstationsActive() - 2; istal >= frAlgo.fFirstCAstation;
         istal--) {  // start with downstream chambers

      for (unsigned int it = 0; it < frAlgo.fTriplets[istal].size(); ++it) {

        ca::Triplet& tr = frAlgo.fTriplets[istal][it];
        tr.SetLevel(0);
        tr.SetFNeighbour(0);
        tr.SetNNeighbours(0);

        if (istal > (frAlgo.GetParameters().GetNstationsActive() - 4)) break;

        unsigned int nNeighbours = frAlgo.fHitNtriplets[tr.GetMHit()];

        unsigned int neighLocation = frAlgo.fHitFirstTriplet[tr.GetMHit()];
        unsigned int neighStation  = frAlgo.TripletId2Station(neighLocation);
        unsigned int neighTriplet  = frAlgo.TripletId2Triplet(neighLocation);

        if (nNeighbours > 0) { assert((int) neighStation == istal + 1 || (int) neighStation == istal + 2); }

        unsigned char level = 0;

        for (unsigned int iN = 0; iN < nNeighbours; ++iN, ++neighTriplet, ++neighLocation) {

          ca::Triplet& neighbour = frAlgo.fTriplets[neighStation][neighTriplet];

          fscal dchi2 = 0.;
          if (!checkTripletMatch(tr, neighbour, dchi2)) continue;

          if (tr.GetFNeighbour() == 0) { tr.SetFNeighbour(neighLocation); }

          tr.SetNNeighbours(neighLocation - tr.GetFNeighbour() + 1);

          const unsigned char neighbourLevel = neighbour.GetLevel();

          if (level <= neighbourLevel) { level = neighbourLevel + 1; }
        }
        tr.SetLevel(level);
      }  // neighbour search

      frAlgo.fMonitorData.IncrementCounter(ECounter::Triplet, frAlgo.fTriplets[istal].size());

    }  // istal
    frAlgo.fMonitorData.StopTimer(ETimer::NeighboringTripletSearch);


    ///====================================================================
    ///=                                                                  =
    ///=        Collect track candidates. CREATE TRACKS                   =
    ///=                                                                  =
    ///====================================================================

    // min level to start triplet. So min track length = min_level+3.
    int min_level = std::min(caIteration.GetMinNhits(), caIteration.GetMinNhitsStation0()) - 3;

    if (frAlgo.fpCurrentIteration->GetTrackFromTripletsFlag()) { min_level = 0; }

    ca::Branch curr_tr;
    ca::Branch new_tr[constants::size::MaxNstations];
    ca::Branch best_tr;

    int ndf = 1;


    // collect consequtive: the longest tracks, shorter, more shorter and so on
    for (int firstTripletLevel = frAlgo.GetParameters().GetNstationsActive() - 3; firstTripletLevel >= min_level;
         firstTripletLevel--) {
      // choose length in triplets number - firstTripletLevel - the maximum possible triplet level among all triplets in the searched candidate

      //  how many levels to check
      int nlevel = (frAlgo.GetParameters().GetNstationsActive() - 2) - firstTripletLevel + 1;

      const unsigned char min_best_l =
        (firstTripletLevel > min_level) ? firstTripletLevel + 2 : min_level + 3;  // loose maximum


      frAlgo.fTrackCandidates.clear();

      for (const auto& h : frAlgo.fWindowHits) {
        frAlgo.fStripToTrack[h.FrontKey()] = -1;
        frAlgo.fStripToTrack[h.BackKey()]  = -1;
      }

      //== Loop over triplets with the required level, find and store track candidates

      for (int istaF = frAlgo.fFirstCAstation;
           istaF <= frAlgo.GetParameters().GetNstationsActive() - 3 - firstTripletLevel; ++istaF) {

        if (--nlevel == 0) break;  //TODO: SG: this is not needed

        for (Tindex itrip = 0; itrip < (Tindex) frAlgo.fTriplets[istaF].size(); ++itrip) {

          ca::Triplet& first_trip = (frAlgo.fTriplets[istaF][itrip]);
          if (frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[first_trip.GetLHit()].FrontKey()]
              || frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[first_trip.GetLHit()].BackKey()]) {
            continue;
          }
          //               ghost suppression !!!
          //

          if (!frAlgo.fpCurrentIteration->GetTrackFromTripletsFlag()) {  // ghost suppression !!!
            int nHits = 3 + first_trip.GetLevel();
            if (frAlgo.fWindowHits[first_trip.GetLHit()].Station() == 0) {
              if (nHits < frAlgo.fpCurrentIteration->GetMinNhitsStation0()) { continue; }
            }
            else {
              if (nHits < frAlgo.fpCurrentIteration->GetMinNhits()) { continue; }
            }
          }

          // Collect triplets, which can start a track with length equal to firstTipletLevel + 3. This cut suppresses
          // ghost tracks, but does not affect the efficiency
          if (first_trip.GetLevel() < firstTripletLevel) { continue; }

          curr_tr.SetChi2(0.);
          curr_tr.SetStation(0);
          curr_tr.ResetHits();
          curr_tr.AddHit(frAlgo.fWindowHits[first_trip.GetLHit()].Id());
          curr_tr.SetChi2(first_trip.GetChi2());

          best_tr = curr_tr;

          CAFindTrack(istaF, best_tr, &first_trip, curr_tr, min_best_l,
                      new_tr);  /// reqursive func to build a tree of possible track-candidates and choose the best

          if (best_tr.NofHits() < firstTripletLevel + 2) continue;  // loose maximum one hit

          if (best_tr.NofHits() < min_level + 3) continue;  // should find all hits for min_level

          ndf = best_tr.NofHits() * 2 - 5;

          // TODO: automatize the NDF calculation

          if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode
              || ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
            ndf = best_tr.NofHits() * 2 - 4;
          }

          best_tr.SetChi2(best_tr.Chi2() / ndf);
          if (frAlgo.fGhostSuppression) {
            if (3 == best_tr.NofHits()) {
              if (!frAlgo.fpCurrentIteration->GetPrimaryFlag() && (istaF != 0))
                continue;  // too /*short*/ non-MAPS track
              if (frAlgo.fpCurrentIteration->GetPrimaryFlag() && (best_tr.Chi2() > 5.0)) continue;
            }
          }
          frAlgo.fTrackCandidates.push_back(best_tr);
          ca::Branch& tr = frAlgo.fTrackCandidates.back();
          tr.SetStation(istaF);
          tr.SetId(frAlgo.fTrackCandidates.size() - 1);
          tr.SetAlive(true);
          if (0) {  // debug
            cout << "track " << frAlgo.fTrackCandidates.size() - 1 << " found, L = " << best_tr.NofHits()
                 << " chi2= " << best_tr.Chi2() << endl;
            cout << " hits: ";
            for (auto hitId : tr.Hits()) {
              cout << frAlgo.GetMcTrackIdForCaHit(hitId) << " ";
            }
            cout << endl;
          }
        }  // itrip
      }    // istaF

      for (ca::Branch& tr : frAlgo.fTrackCandidates) {
        tr.SetAlive(false);
      }

      bool repeatCompetition = true;
      int iComp              = 0;
      for (iComp = 0; (iComp < 100) && repeatCompetition; ++iComp) {

        // == Loop over track candidates and mark their strips

        repeatCompetition = false;

        for (ca::Branch& tr : frAlgo.fTrackCandidates) {
          if (tr.IsAlive()) { continue; }
          for (auto& hitId : tr.Hits()) {
            const ca::Hit& h = frAlgo.fInputData.GetHit(hitId);
            bool isAlive     = true;
            {  // front  strip
              auto& stripF = (frAlgo.fStripToTrack)[h.FrontKey()];
              if ((stripF >= 0) && (stripF != tr.Id())) {  // strip is used by other candidate
                const auto& other = frAlgo.fTrackCandidates[stripF];
                if (!other.IsAlive() && tr.IsBetterThan(other)) { stripF = tr.Id(); }
                else {
                  isAlive = false;
                }
              }
              else {
                stripF = tr.Id();
              }
              if (!isAlive) { break; }
            }

            {  // back strip
              auto& stripB = (frAlgo.fStripToTrack)[h.BackKey()];
              if ((stripB >= 0) && (stripB != tr.Id())) {  // strip is used by other candidate
                const auto& other = frAlgo.fTrackCandidates[stripB];
                if (!other.IsAlive() && tr.IsBetterThan(other)) { stripB = tr.Id(); }
                else {
                  isAlive = false;
                }
              }
              else {
                stripB = tr.Id();
              }
              if (!isAlive) { break; }
            }
          }  // loop over hits
        }    // itrack

        // == Check if some suppressed candidates are still alive

        for (ca::Branch& tr : frAlgo.fTrackCandidates) {
          if (tr.IsAlive()) { continue; }

          tr.SetAlive(true);
          for (int iHit = 0; tr.IsAlive() && (iHit < (int) tr.Hits().size()); ++iHit) {
            const ca::Hit& h = frAlgo.fInputData.GetHit(tr.Hits()[iHit]);
            tr.SetAlive(tr.IsAlive() && ((frAlgo.fStripToTrack)[h.FrontKey()] == tr.Id())
                        && ((frAlgo.fStripToTrack)[h.BackKey()] == tr.Id()));
          }

          if (!tr.IsAlive()) {  // release strips
            for (auto hitId : tr.Hits()) {
              const ca::Hit& h = frAlgo.fInputData.GetHit(hitId);
              if (frAlgo.fStripToTrack[h.FrontKey()] == tr.Id()) { frAlgo.fStripToTrack[h.FrontKey()] = -1; }
              if (frAlgo.fStripToTrack[h.BackKey()] == tr.Id()) { frAlgo.fStripToTrack[h.BackKey()] = -1; }
            }
          }
          else {
            repeatCompetition = true;
          }
        }  // itrack

      }  // competitions

      // cout << " N Competitions: " << iComp << endl;

      // ==

      for (Tindex iCandidate = 0; iCandidate < (Tindex) frAlgo.fTrackCandidates.size(); ++iCandidate) {
        ca::Branch& tr = frAlgo.fTrackCandidates[iCandidate];
        if (!tr.IsAlive()) continue;
        if (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
          if (tr.NofHits() <= 3) { continue; }
        }
        else if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode) {
          if (tr.NofHits() < 3) { continue; }
        }
        else {
          if (tr.NofHits() < 3) { continue; }
        }
        if (frAlgo.fpCurrentIteration->GetExtendTracksFlag()) {
          if (tr.NofHits() < frAlgo.GetParameters().GetNstationsActive()) { frAlgo.fTrackExtender.ExtendBranch(tr); }
        }

        for (auto iHit : tr.Hits()) {
          const ca::Hit& hit = frAlgo.fInputData.GetHit(iHit);

          /// used strips are marked

          frAlgo.fvHitKeyFlags[hit.FrontKey()] = 1;
          frAlgo.fvHitKeyFlags[hit.BackKey()]  = 1;

          frAlgo.fSliceRecoHits.push_back(iHit);
        }
        Track t;
        t.fNofHits = tr.NofHits();
        frAlgo.fSliceRecoTracks.push_back(t);
        if (0) {  // SG debug
          cout << "store track " << iCandidate << " chi2= " << tr.Chi2() << endl;
          cout << " hits: ";
          for (unsigned int ih = 0; ih < tr.Hits().size(); ih++) {
            cout << frAlgo.GetMcTrackIdForCaHit(tr.Hits()[ih]) << " ";
          }
          cout << '\n';
        }

      }  // tracks

    }  // firstTripletLevel


    // suppress strips of suppressed hits
    for (unsigned int ih = 0; ih < frAlgo.fWindowHits.size(); ih++) {
      if (frAlgo.fIsWindowHitSuppressed[ih]) {
        const ca::Hit& hit                   = frAlgo.fWindowHits[ih];
        frAlgo.fvHitKeyFlags[hit.FrontKey()] = 1;
        frAlgo.fvHitKeyFlags[hit.BackKey()]  = 1;
      }
    }
  }  // ---- Loop over Track Finder iterations: END -----------------------------------------------------------//

  // Fit tracks
  frAlgo.fTrackFitter.FitCaTracks();

  // Merge clones
  frAlgo.fCloneMerger.Exec(frAlgo.fSliceRecoTracks, frAlgo.fSliceRecoHits);
}


/** *************************************************************
     *                                                              *
     *     The routine performs recursive search for tracks         *
     *                                                              *
     *     I. Kisel                                    06.03.05     *
     *     I.Kulakov                                    2012        *
     *                                                              *
     ****************************************************************/

void TrackFinderWindow::CAFindTrack(int ista, ca::Branch& best_tr, const ca::Triplet* curr_trip, ca::Branch& curr_tr,
                                    unsigned char min_best_l, ca::Branch* new_tr)
/// recursive search for tracks
/// input: @ista - station index, @&best_tr - best track for the privious call
/// output: @&NCalls - number of function calls
{

  if (curr_trip->GetLevel() == 0)  // the end of the track -> check and store
  {
    // -- finish with current track
    // add rest of hits
    const ca::HitIndex_t& ihitm = curr_trip->GetMHit();
    const ca::HitIndex_t& ihitr = curr_trip->GetRHit();

    if (!(frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[ihitm].FrontKey()]
          || frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[ihitm].BackKey()])) {

      //        curr_tr.Hits.push_back(fGridHitIds[ihitm]);

      curr_tr.AddHit(frAlgo.fWindowHits[ihitm].Id());
    }

    if (!(frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[ihitr].FrontKey()]
          || frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[ihitr].BackKey()])) {

      //curr_tr.Hits.push_back(fGridHitIds[ihitr]);
      curr_tr.AddHit(frAlgo.fWindowHits[ihitr].Id());
    }

    //if( curr_tr.NofHits() < min_best_l - 1 ) return; // suppouse that only one hit can be added by extender
    // TODO: SZh 21.08.2023: Replace hard-coded value with a parameter
    int ndf = curr_tr.NofHits() * 2 - 5;

    if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode
        || ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
      ndf = curr_tr.NofHits() * 2 - 4;
    }

    if (curr_tr.Chi2() > frAlgo.fTrackChi2Cut * ndf) { return; }


    // -- select the best
    if ((curr_tr.NofHits() > best_tr.NofHits())
        || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
      best_tr = curr_tr;
    }

    return;
  }
  else  //MEANS level ! = 0
  {
    int N_neighbour = (curr_trip->GetNNeighbours());

    for (Tindex in = 0; in < N_neighbour; in++) {

      unsigned int ID = curr_trip->GetFNeighbour() + in;

      unsigned int Station = frAlgo.TripletId2Station(ID);
      unsigned int Triplet = frAlgo.TripletId2Triplet(ID);

      const ca::Triplet& new_trip = frAlgo.fTriplets[Station][Triplet];

      fscal dchi2 = 0.;
      if (!checkTripletMatch(*curr_trip, new_trip, dchi2)) continue;

      if (frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[new_trip.GetLHit()].FrontKey()]
          || frAlgo.fvHitKeyFlags[frAlgo.fWindowHits[new_trip.GetLHit()].BackKey()]) {  //hits are used
        //  no used hits allowed -> compare and store track
        if ((curr_tr.NofHits() > best_tr.NofHits())
            || ((curr_tr.NofHits() == best_tr.NofHits()) && (curr_tr.Chi2() < best_tr.Chi2()))) {
          best_tr = curr_tr;
        }
      }
      else {  // hit is not used: add the left hit from the new triplet to the current track

        unsigned char new_L = curr_tr.NofHits() + 1;
        fscal new_chi2      = curr_tr.Chi2() + dchi2;


        if (0) {  //SGtrd2d debug!!
          int mc01 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetLHit());
          int mc02 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetMHit());
          int mc03 = frAlgo.GetMcTrackIdForWindowHit(curr_trip->GetRHit());
          int mc11 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetLHit());
          int mc12 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetMHit());
          int mc13 = frAlgo.GetMcTrackIdForWindowHit(new_trip.GetRHit());
          if ((mc01 == mc02) && (mc02 == mc03)) {
            cout << " sta " << ista << " mc0 " << mc01 << " " << mc02 << " " << mc03 << " mc1 " << mc11 << " " << mc12
                 << " " << mc13 << " chi2 " << curr_tr.Chi2() / (2 * (curr_tr.NofHits() + 2) - 4) << " new "
                 << new_chi2 / (2 * (new_L + 2) - 4) << endl;
            cout << "   hits " << curr_trip->GetLHit() << " " << curr_trip->GetMHit() << " " << curr_trip->GetRHit()
                 << " " << new_trip.GetLHit() << endl;
          }
        }

        int ndf = 2 * (new_L + 2) - 5;

        if (ca::Framework::TrackingMode::kGlobal == frAlgo.fTrackingMode) {  //SGtrd2d!!!
          ndf = 2 * (new_L + 2) - 4;
        }
        else if (ca::Framework::TrackingMode::kMcbm == frAlgo.fTrackingMode) {
          ndf = 2 * (new_L + 2) - 4;
        }
        else {
          ndf = new_L;  // TODO: SG:  2 * (new_L + 2) - 5
        }

        if (new_chi2 > frAlgo.fTrackChi2Cut * ndf) continue;

        // add new hit
        new_tr[ista] = curr_tr;
        new_tr[ista].AddHit(frAlgo.fWindowHits[new_trip.GetLHit()].Id());

        const int new_ista = ista + new_trip.GetMSta() - new_trip.GetLSta();
        new_tr[ista].SetChi2(new_chi2);

        CAFindTrack(new_ista, best_tr, &new_trip, new_tr[ista], min_best_l, new_tr);
      }  // add triplet to track
    }    // for neighbours
  }      // level = 0
}
