/* Copyright (C) 2023 Facility for Antiproton and Ion Research in Europe, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Dominik Smith [committer] */

#include "Hitfind.h"

#include <chrono>

#include "log.hpp"
#include "util/TimingsFormat.h"

using namespace std;
using fles::Subsystem;

namespace cbm::algo::tof
{
  // -----   Execution   -------------------------------------------------------
  Hitfind::resultType Hitfind::operator()(gsl::span<CbmTofDigi> digiIn)
  {
    xpu::push_timer("TofHitfind");
    xpu::t_add_bytes(digiIn.size_bytes());

    // --- Output data
    resultType result = {};

    auto& clusterTs = std::get<0>(result);
    auto& monitor   = std::get<1>(result);

    // Loop over the digis array and store the Digis in separate vectors for
    // each RPC modules
    for (int32_t idigi = 0; idigi < digiIn.size(); idigi++) {
      CbmTofDigi* pDigi = &(digiIn[idigi]);

      // These are doubles in the digi class
      const double SmType = pDigi->GetType();
      const double Sm     = pDigi->GetSm();
      const double Rpc    = pDigi->GetRpc();
      const int NbRpc     = fNbRpc[SmType];
      if (SmType >= fNbSm.size() || Sm * NbRpc + Rpc >= fNbSm[SmType] * NbRpc) {
        // Error already counted for monitoring during Digis calibration, so just discard here
        continue;
      }
      fStorDigi[SmType][Sm * NbRpc + Rpc].push_back(std::make_pair(pDigi, idigi));
    }

    std::vector<Hit> clustersFlat;
    std::vector<size_t> rpcSizes;   // nClusters per RPC
    std::vector<u32> rpcAddresses;  // RPC addresses

    // ---  RPC loop
    for (uint32_t SmType = 0; SmType < fNbSm.size(); SmType++) {
      const int NbRpc = fNbRpc[SmType];
      const int NbSm  = fNbSm[SmType];
      for (uint32_t Sm = 0; Sm < NbSm; Sm++) {
        for (uint32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          // Get digis
          std::vector<std::pair<CbmTofDigi*, int32_t>>& digiExp = fStorDigi[SmType][Sm * NbRpc + Rpc];

          // Build clusters
          std::vector<cbm::algo::tof::Hit> clusters = fAlgo[SmType][Sm * NbRpc + Rpc](digiExp);

          // Store hw address of partition
          // TODO: is this correct? Does this address make sense?
          u32 rpcAddress = CbmTofAddress::GetUniqueAddress(Sm, Rpc, 0, 0, SmType);
          rpcAddresses.push_back(rpcAddress);

          // store partition size
          rpcSizes.push_back(clusters.size());

          // Append clusters to output
          clustersFlat.insert(clustersFlat.end(), clusters.begin(), clusters.end());

          // Clear digi storage
          digiExp.clear();
        }
      }
    }

    // Monitoring
    monitor.fTime     = xpu::pop_timer();
    monitor.fNumDigis = digiIn.size();
    monitor.fNumHits  = clustersFlat.size();

    // Create ouput vector
    clusterTs = PartitionedVector(std::move(clustersFlat), rpcSizes, rpcAddresses);

    return result;
  }
  // ----------------------------------------------------------------------------


  // -----   Initialisation   ---------------------------------------------------
  void Hitfind::Init()
  {
    //Number of SMs per super module type
    fNbSm = fTofConfig.NbSm;

    //Number of RPCs per super module type
    fNbRpc = fTofConfig.NbRpc;

    //Prepare digi storage
    fStorDigi.resize(fNbSm.size());

    // Create one algorithm per RPC for TOF and configure it with parametersa
    for (int32_t SmType = 0; SmType < fNbSm.size(); SmType++) {

      int32_t NbSm  = fNbSm[SmType];
      int32_t NbRpc = fNbRpc[SmType];
      fStorDigi[SmType].resize(NbSm * NbRpc);

      for (int32_t Sm = 0; Sm < NbSm; Sm++) {
        for (int32_t Rpc = 0; Rpc < NbRpc; Rpc++) {

          std::unique_ptr<cbm::algo::tof::ClusterizerRpcPar> par(new cbm::algo::tof::ClusterizerRpcPar());

          HitfindSetup::Rpc rpcPar = fTofConfig.rpcs[SmType][Sm * NbRpc + Rpc];
          par->fDeadStrips         = rpcPar.deadStrips;
          par->fPosYMaxScal        = rpcPar.posYMaxScal;
          par->fdMaxTimeDist       = rpcPar.maxTimeDist;
          par->fdMaxSpaceDist      = rpcPar.maxSpaceDist;
          par->fSigVel             = rpcPar.sigVel;
          par->fCPTOffYBinWidth    = rpcPar.CPTOffYBinWidth;
          par->fCPTOffY            = rpcPar.CPTOffY;
          par->fCPTOffYRange       = rpcPar.CPTOffYRange;
          par->fTimeRes            = rpcPar.timeRes;

          //geometry information, can in principle be done channel-wise but not needed for now
          const double* tra_ptr = rpcPar.cell.translation.data();
          const double* rot_ptr = rpcPar.cell.rotation.data();

          //channel parameters
          int32_t NbChan = rpcPar.chanPar.size();
          par->fChanPar.resize(NbChan);

          for (int32_t Ch = 0; Ch < NbChan; Ch++) {
            HitfindSetup::Channel chanPar   = rpcPar.chanPar[Ch];
            par->fChanPar[Ch].address       = chanPar.address;
            par->fChanPar[Ch].cell.pos      = ROOT::Math::XYZVector(tra_ptr[0], tra_ptr[1], tra_ptr[2]);
            par->fChanPar[Ch].cell.rotation = ROOT::Math::Rotation3D(&rot_ptr[0], &rot_ptr[9]);
            par->fChanPar[Ch].cell.sizeX    = rpcPar.cell.sizeX;
            par->fChanPar[Ch].cell.sizeY    = rpcPar.cell.sizeY;
          }
          fAlgo[SmType][Sm * NbRpc + Rpc].SetParams(std::move(par));
        }
      }
    }
    L_(info) << "--- Configured hitfinder algorithms for TOF.";
    L_(info) << "==================================================";
  }
  // ----------------------------------------------------------------------------

}  // namespace cbm::algo::tof
